/*
** The guts of these two functions, grabImageCB & grabRegionImageCB, were
** taken from "xmag" by Jim Fulton...
**
**
** xmag - simple magnifying glass application
**
** $XConsortium: xmag.c,v 1.19 88/10/13 16:52:05 jim Exp $
**
** Copyright 1988 Massachusetts Institute of Technology
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation, and that the name of M.I.T. not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  M.I.T. makes no representations about the
** suitability of this software for any purpose.  It is provided "as is"
** without express or implied warranty.
**
** Author:  Jim Fulton, MIT X Consortium
*/


#include "pixt.h"
#define NPOINTS 49

void grabImageCB(w,client,call)
	Widget w;
	caddr_t client, call;
{
	int width=NumCols, height=NumRows;
	/* assumes the existence of a global variable display */
	XImage *xi;
	register int x, y;
	int screen = DefaultScreen(display),
	    dw = DisplayWidth (display, screen), 
	    dh = DisplayHeight (display, screen),
	    root_x, root_y, win_x, win_y,
	    xoff = width / 2, yoff = height / 2,
	    lastx, lasty;
	Window root_window, child_window, root = RootWindow (display, screen);
	XPoint box[NPOINTS];
	unsigned long mask;
	Bool done;
	GC invGC = (GC)NULL;
	XGCValues gcv;
	Cursor invCursor;
	XWindowAttributes winatts;

        invCursor = XCreateFontCursor (display, XC_crosshair);
	gcv.function = GXxor;
        gcv.subwindow_mode = IncludeInferiors;
        gcv.foreground = (unsigned long) 1;
        gcv.plane_mask = (unsigned long) 1;

	invGC = XCreateGC (display, win, (GCFunction | GCSubwindowMode |
                                       GCForeground | GCPlaneMask), &gcv);

	XmUpdateDisplay(toplevel);
	XFlush(display);
	XGrabServer (display);
	XSync (display, 1);
	if (XGrabPointer (display, root, False, ButtonPressMask,
	    GrabModeAsync, GrabModeAsync, None,
	    invCursor, CurrentTime) != GrabSuccess) {
		fprintf (stderr, "unable to grab pointer\n");
		exit (0);
	}
#define FAROFFSCREEN -9999

	for (done = False, lastx = lasty = FAROFFSCREEN; !done; ) {
		if (XPending (display) != 0) {
			XEvent event;
			XNextEvent (display, &event);
			switch (event.type) {
			case ButtonPress:
				root_x = event.xbutton.x_root;
				root_y = event.xbutton.y_root;
				done = True;
				/* get the attributes of the window in which
				 * this blessed event occurred
				 */
				XGetWindowAttributes(display,
					event.xbutton.window,&winatts);
				break;
			default:
                		continue;
			} 
		} 
		else if (!XQueryPointer (display, root, &root_window, 
				&child_window,
				&root_x, &root_y, 
				&win_x, &win_y, &mask)) {
            		fprintf (stderr, 
				"unable to QueryPointer on root window.\n");
			exit(0);
		}
	        if (root_x != lastx || root_y != lasty) {
            		register int i;

			/* clip to screen */
			x = root_x - xoff;
			y = root_y - yoff;
			if (x < 0) x = 0;
			if (y < 0) y = 0;
			if (x + width >= dw) x = dw - 1 - width;
			if (y + height >= dh) y = dh - 1 - height;

			/* fill in the outline box */
			for (i = 0; i < NPOINTS; i += 4) {
				box[i].x = x; 
				box[i].y = y;  /*ul*/
				if (i == (NPOINTS - 1)) break;
				box[i+1].x = x + width; 
				box[i+1].y = y; /*ur*/
				box[i+2].x = box[i+1].x; 
				box[i+2].y = y + height;  /* l.r. */
				box[i+3].x = x; 
				box[i+3].y = box[i+2].y;  /* l.l. */
			}
			lastx = root_x;
			lasty = root_y;
		}
			
		XDrawLines (display,root,invGC,box,NPOINTS,CoordModeOrigin);
	}
	XUngrabPointer (display, CurrentTime);
	XFlush (display);
	xi = XGetImage(display, root, x, y, width, height,
                       AllPlanes, ZPixmap);
	XUngrabServer(display);
	XSync(display,0);
	{
	int row, col, i;
	for (row=0;row<height;row++)
		for (col=0;col<width;col++) {
			i = row*width+col;
			PixColors[i] = XGetPixel(xi,col,row);
			XSetForeground(display,paintGC,PixColors[i]);
			XFillRectangle(display,win,paintGC,
					ColSegs[col].x1+1,
					RowSegs[row].y1+1,
					CELLSIZE,CELLSIZE);
		}
	}
	redrawPixmap(0,0,xi->width,xi->height);
}

void grabRegionImageCB(w,client,call)
	Widget w;
	caddr_t client, call;
{
	int width=ActiveRegion.width+1, height=ActiveRegion.height+1;
	XImage *xi;
	register int x, y;
	int screen = DefaultScreen(display),
	    dw = DisplayWidth (display, screen), 
	    dh = DisplayHeight (display, screen),
	    root_x, root_y, win_x, win_y,
	    xoff = width / 2, yoff = height / 2,
	    lastx, lasty;
	Window root_window, child_window, root = RootWindow (display, screen);
	XPoint box[NPOINTS];
	unsigned long mask;
	Bool done;
	GC invGC = (GC)NULL;
	XGCValues gcv;
	Cursor invCursor;
	XWindowAttributes winatts;

	if (width == 0 || height == 0)
		return;
        invCursor = XCreateFontCursor (display, XC_crosshair);
	gcv.function = GXxor;
        gcv.subwindow_mode = IncludeInferiors;
        gcv.foreground = (unsigned long) 1;
        gcv.plane_mask = (unsigned long) 1;

	invGC = XCreateGC (display, win, (GCFunction | GCSubwindowMode |
                                       GCForeground | GCPlaneMask), &gcv);
	XmUpdateDisplay(toplevel);
	XFlush(display);
	XGrabServer (display);
	XSync (display, 1);
	if (XGrabPointer (display, root, False, ButtonPressMask,
	    GrabModeAsync, GrabModeAsync, None,
	    invCursor, CurrentTime) != GrabSuccess) {
		fprintf (stderr, "unable to grab pointer\n");
		exit (0);
	}
#define FAROFFSCREEN -9999

	for (done = False, lastx = lasty = FAROFFSCREEN; !done; ) {
		if (XPending (display) != 0) {
			XEvent event;
			XNextEvent (display, &event);
			switch (event.type) {
			case ButtonPress:
				root_x = event.xbutton.x_root;
				root_y = event.xbutton.y_root;
				done = True;
				/* get the attributes of the window in which
				 * this blessed event occurred
				 */
				XGetWindowAttributes(display,
					event.xbutton.window,&winatts);
				break;
			default:
                		continue;
			} 
		} 
		else if (!XQueryPointer (display, root, &root_window, 
				&child_window,
				&root_x, &root_y, 
				&win_x, &win_y, &mask)) {
            		fprintf (stderr, 
				"unable to QueryPointer on root window.\n");
			exit(0);
		}
	        if (root_x != lastx || root_y != lasty) {
            		register int i;

			/* clip to screen */
			x = root_x - xoff;
			y = root_y - yoff;
			if (x < 0) x = 0;
			if (y < 0) y = 0;
			if (x + width >= dw) x = dw - 1 - width;
			if (y + height >= dh) y = dh - 1 - height;

			/* fill in the outline box */
			for (i = 0; i < NPOINTS; i += 4) {
				box[i].x = x; 
				box[i].y = y;  /*ul*/
				if (i == (NPOINTS - 1)) break;
				box[i+1].x = x + width; 
				box[i+1].y = y; /*ur*/
				box[i+2].x = box[i+1].x; 
				box[i+2].y = y + height;  /* l.r. */
				box[i+3].x = x; 
				box[i+3].y = box[i+2].y;  /* l.l. */
			}
			lastx = root_x;
			lasty = root_y;
		}
			
		XDrawLines (display,root,invGC,box,NPOINTS,CoordModeOrigin);
	}
	XUngrabPointer (display, CurrentTime);
	XFlush (display);
	xi = XGetImage(display, root, x, y, width, height,
                       AllPlanes, ZPixmap);
	XUngrabServer(display);
	XSync(display,0);
	{
	int row, col, i;
	int x, y;
	for (row=ActiveRegion.row,y=0;
	     row<ActiveRegion.height+ActiveRegion.row+1;
	     row++,y++)
		for (col=ActiveRegion.col,x=0;
		     col<ActiveRegion.width+ActiveRegion.col+1;
		     col++,x++) {
			i = row*NumCols+col;
			PixColors[i] = XGetPixel(xi,x,y);
			XSetForeground(display,paintGC,PixColors[i]);
			XFillRectangle(display,win,paintGC,
					ColSegs[col].x1+1,
					RowSegs[row].y1+1,
					CELLSIZE,CELLSIZE);
		}
		XPutImage(display,pixwin,paintGC,xi,0,0,
				ActiveRegion.col,
				ActiveRegion.row,
				ActiveRegion.width,
				ActiveRegion.height);
	}
}
