/* Copyright 1989 GROUPE BULL -- See licence conditions in file COPYRIGHT */
/*****************************************************************************
 ****  Create utility for XPM file format... [ X PixMap ]
 ****
 ****  Pixmap XCreatePixmapFromData(dpy, d, cmap, w, h, depth, n, c, col, pix)
 ****
 ****  original:  Daniel Dardailler - Bull RC (89/02/22) daniel@mirsa.inria.fr
 ****  contributors:
 ****             James Bash - AT&T Bell Labs            jmb@attunix.att.com
 *****************************************************************************
 ****  Version 1.2:  Improved storage mgmt & fatal error handling...
 ****  Version 1.1:  extended chars_per_pixel support... [ Read, Create ]
-------------------------------------------------------------------------------
WHO:	Richard Hess			CORP:	Consilium
TITLE:	Project Manager			VOICE:	[415] 691-6342
      [ User Interface Services ]	USNAIL:	640 Clyde Court
UUCP:	...!uunet!cimshop!rhess			Mountain View, CA 94043
-------------------------------------------------------------------------------
 */

#include <stdio.h>

#ifdef VMS
#include "decw$include:Xlib.h"
#else
#include <X11/Xlib.h>
#endif

#include "xpm.h"          /* PixmapOpenFailed, PixmapSuccess .. MAXPRINTABLE */

/**[ XCreatePixmapFromData ]**************************************************
 *
 *  This function allows you to include in your C program (using #include)
 *  a pixmap file in XPM format that was written by XWritePixmapFile
 * (THIS VERSION SUPPORTS : '_chars_per_pixel 1', '_chars_per_pixel 2')
 *****************************************************************************/

Pixmap XCreatePixmapFromData(display, d, colormap, width, height, depth, 
			     ncolors, chars_per_pixel, colors, pixels)
     Display *display;
     Drawable d;
     Colormap colormap ;
     unsigned int width, height;       
     int depth ;   
     unsigned int ncolors ;
     unsigned int chars_per_pixel ;
     char ** colors ;           /* array of colormap entries  "cc","#RRGGBB" */
     char ** pixels ;           /* array of pixels lines    "cc..cc00cccc.." */
{
  Pixmap pixmap ;
  GC Gc = NULL;
  XGCValues xgcv;
  XImage *xim = NULL;

  CmapEntry * cmap = NULL;
  int * Tpixel = NULL;

  char c1, c2, c;
  int red, green, blue ;
  XColor xcolor ;
  int i,j,p;
  
#ifdef DEBUG
#define DEBUGG fprintf
#else
#define DEBUGG
#endif

#undef RETURN
#define RETURN(code, flag) \
  { if (cmap)   free (cmap); \
    if (Tpixel) free (Tpixel); \
    if (Gc)     XFreeGC(display, Gc); \
    if (xim)    XDestroyImage(xim); \
    if (flag && pixmap) XFreePixmap(display, pixmap); \
    return (code); }

  if (ncolors > (MAXPRINTABLE*MAXPRINTABLE)) {
    DEBUGG (stderr,"XPM: Too many different colors, version 1\n");
    RETURN (PixmapFileInvalid, 0);
  }

  if ((chars_per_pixel < 1) || (chars_per_pixel > 2)) {
    DEBUGG (stderr,"XPM: version 1.1 handles only 1 or 2 chars_per_pixel\n");
    RETURN (PixmapFileInvalid, 0);
  }

  /* now we construct cmap and Tpixel from colors array parameter */
  cmap = (CmapEntry *) malloc(ncolors*sizeof(CmapEntry)) ;
  Tpixel = (int *) malloc(ncolors*sizeof(int)) ;
  if ((cmap == NULL) || (Tpixel == NULL)) {
    RETURN (PixmapNoMemory, 0);
  }

  if (!colors) {
    DEBUGG (stderr,"XPM: colors not defined\n");
    RETURN (PixmapFileInvalid, 0);
  }

  switch (chars_per_pixel)
    {
    case 1:
      for (i=0; i<2*ncolors ; i+=2)
	{
	  if (sscanf(colors[i],"%c", &c1) != 1) {
	    DEBUGG (stderr,
		    "XPM: bad cmap entry, must be '\"c\" , \"colordef\",'\n"); 
	    RETURN (PixmapFileInvalid, 0);
	  }
	  if (strchr(printable,c1))
	    {
	      cmap[i/2].cixel.c1 = c1 ;
	      if (!XParseColor(display,colormap,colors[i+1],&xcolor)) {
		DEBUGG (stderr,
			"XPM: bad colordef spec, #RGB or colorname\n");
		RETURN (PixmapFileInvalid, 0);
	      }
	      XAllocColor(display,colormap,&xcolor);
	      Tpixel[i/2] = xcolor.pixel ;
	    } else {
	      DEBUGG (stderr,"XPM: bad cixel value, must be printable\n");
	      RETURN (PixmapFileInvalid, 0);
	    }
	}
      break;
    case 2:
      for (i=0; i<2*ncolors ; i+=2)
	{
	  if (sscanf(colors[i],"%c%c", &c1,&c2) != 2) {
	    DEBUGG (stderr,
		   "XPM: bad cmap entry, must be '\"cC\" , \"colordef\",'\n"); 
	    RETURN (PixmapFileInvalid, 0);
	  }
	  if ((strchr(printable,c1)) &&
	      (strchr(printable,c2))) {
	    cmap[i/2].cixel.c1 = c1 ;
	    cmap[i/2].cixel.c2 = c2 ;
	    if (!XParseColor(display,colormap,colors[i+1],&xcolor)) {
	      DEBUGG (stderr,
		      "XPM: bad colordef spec, #RGB or colorname\n");
	      RETURN (PixmapFileInvalid, 0);
	    }
	    XAllocColor(display,colormap,&xcolor);
	    Tpixel[i/2] = xcolor.pixel ;
	  } else {
	    DEBUGG (stderr,"XPM: bad cixel value, must be printable\n");
	    RETURN (PixmapFileInvalid, 0);
	  }
	}
      break;
    }
  
  pixmap = XCreatePixmap(display,d,width,height,depth);
  Gc = XCreateGC(display,pixmap,0,&xgcv);
  xim = XGetImage(display,pixmap,0,0,width,height,AllPlanes,ZPixmap);
  
  if (!pixels) {
    DEBUGG (stderr,"XPM: pixels not defined\n");
    RETURN (PixmapFileInvalid, 1);
  }

  j = 0 ;
  while (j < height)
    {  
      if (strlen(pixels[j]) != (chars_per_pixel*width)) {
	DEBUGG (stderr,"XPM: bad pixmap line length %d\n",strlen(pixels[j]));
	RETURN (PixmapFileInvalid, 1);
      }
      switch (chars_per_pixel)
	{
	case 1:
	  for (i=0; i< (width) ; i++)
	    {
	      c1 = pixels[j][i] ;
	      for (p = 0 ; p < ncolors ; p++)
		if (cmap[p].cixel.c1 == c1) break ;
	      if (p == ncolors) {
		DEBUGG (stderr,
			"XPM: cixel \"%c\" not in previous cmap\n",c1);
		RETURN (PixmapFileInvalid, 1);
	      }
	      XPutPixel(xim,i,j,Tpixel[p]);
	    }
	  break;
	case 2:
	  for (i=0; i< (2*width) ; i+=2)
	    {
	      c1 = pixels[j][i] ;
	      c2 = pixels[j][i+1] ;
	      for (p = 0 ; p < ncolors ; p++)
		if ((cmap[p].cixel.c1 == c1)&&(cmap[p].cixel.c2 == c2)) break ;
	      if (p == ncolors) {
		DEBUGG (stderr,
			"XPM: cixel \"%c%c\" not in previous cmap\n",c1,c2);
		RETURN (PixmapFileInvalid, 1);
	      }
	      XPutPixel(xim,i/2,j,Tpixel[p]);
	    }
	  break;
	}
      j++ ;
    }
  
  XPutImage(display,pixmap,Gc,xim,0,0,0,0,width,height);

  RETURN (pixmap, 0);
}

/****<eof>********************************************************************/
