/* Copyright 1989 GROUPE BULL -- See licence conditions in file COPYRIGHT */
/*****************************************************************************
 ****  Write utility for XPM file format... [ X PixMap ]
 ****
 ****  int XWritePixmapFile(dpy, cmap, filename, pixmap, w, h)
 ****
 ****  original:  Daniel Dardailler - Bull RC (89/02/22) daniel@mirsa.inria.fr
 ****  contributors:
 ****             James Bash - AT&T Bell Labs            jmb@attunix.att.com
 *****************************************************************************
 ****  Version 1.1:  Improved file & storage mgmt...
-------------------------------------------------------------------------------
WHO:	Richard Hess			CORP:	Consilium
TITLE:	Project Manager			VOICE:	[415] 691-6342
      [ User Interface Services ]	USNAIL:	640 Clyde Court
UUCP:	...!uunet!cimshop!rhess			Mountain View, CA 94043
-------------------------------------------------------------------------------
 */

#include <stdio.h>

#ifdef VMS
#include "decw$include:Xlib.h"
#else
#include <X11/Xlib.h>
#endif

#include "xpm.h"          /* PixmapOpenFailed, PixmapSuccess .. MAXPRINTABLE */

#ifdef SYSV
#define rename(from, to)	(link(from, to) ? -1 : (unlink(from) ? -1 : 0))
#endif /* SYSV */

#define MAX_LINE_LENGTH		512

static char * StripName();
static char * BackupName();
static char * TmpName();

/**[ XWritePixmapFile ]*******************************************************
 *
 *  Write a X Color Pixmap through a specified colormap
 * (THIS VERSION PRODUCES 1 CHARS PER PIXEL FORMAT : '_chars_per_pixel 1')
 * (THIS VERSION PRODUCES 2 CHARS PER PIXEL FORMAT : '_chars_per_pixel 2')
 *****************************************************************************/

int XWritePixmapFile(display,
		     colormap, filename, pixmap, width, height)
     Display * display;
     Colormap  colormap ;
     char *filename;
     Pixmap pixmap;
     unsigned int width, height;
{
   XImage *image = NULL;         /* client image */
   
   FILE *fstream = NULL;
   char *stripname = NULL; 
   char *backupname = NULL;
   char *tmpname = NULL;

   int ncolors = 0 ;             /* number of different colors */

   CmapEntry * cmap = NULL;      /* indexed by ncolors, 
				    dynamically allocated by width */

   Cixel * Map = NULL;           /* should be [width][height] */
   int i,j,p ;

   XColor * colors = NULL;

   int MaxCixel ;               /* theorytical max */
   int cpp;

#undef RETURN
#define RETURN(code) \
   { if (stripname) free (stripname); \
     if (backupname) free (backupname); \
     if (tmpname) free (tmpname); \
     if (cmap) free (cmap); \
     if (Map) free (Map); \
     if (colors) free (colors); \
     if (image) XDestroyImage(image); \
     if (fstream) { fclose (fstream); unlink (tmpname); } \
     return (code); }

   if (!filename) { RETURN(PixmapOpenFailed); }

   if ((stripname = StripName(filename)) == NULL ||
       (backupname = BackupName(filename)) == NULL ||
       (tmpname = TmpName(filename)) == NULL) {
     RETURN (PixmapNoMemory);
   }

   if (!(fstream = fopen(tmpname, "w"))) {
     RETURN(PixmapOpenFailed);            
   }

   /* Convert pixmap to an image in client memory */
   image = XGetImage(display, pixmap, 0,0,width, height, AllPlanes, ZPixmap);

   /* calcul of ncolors, cmap, Map in a while */
   colors = (XColor *) calloc(width,sizeof(XColor));
   cmap = (CmapEntry *) malloc(MAXPRINTABLE*sizeof(CmapEntry)); 
   Map = (Cixel *) malloc(sizeof(Cixel)*width*height);

   if ((colors == NULL) || (cmap == NULL) || (Map == NULL)) {
     RETURN (PixmapNoMemory);
   }

   MaxCixel = (MAXPRINTABLE*MAXPRINTABLE)-1 ;

   for (j=0 ; j<height ; j++) {
     for (i=0; i<width ; i++) colors[i].pixel = XGetPixel(image,i,j) ;
     /* ask server for rgb values line by line */
     XQueryColors(display,colormap,colors,width);  /* return : colors */
     for (i=0; i<width ; i++) {
       for (p=0; p<ncolors; p++)
	 if ((colors[i].red   == cmap[p].red) &&
	     (colors[i].green == cmap[p].green) &&
	     (colors[i].blue  == cmap[p].blue)) {  /* already present */
	   *(Map + j*width + i) = cmap[p].cixel ;
	   break ;                     /* stop with p < ncolors */
	 }
       if (p == ncolors) {             /* new rgb value in cmap */
	 cmap[ncolors].red = colors[i].red ;
	 cmap[ncolors].green = colors[i].green ;
	 cmap[ncolors].blue = colors[i].blue ;
	 cmap[ncolors].cixel.c1 = printable[ncolors % MAXPRINTABLE];
	 cmap[ncolors].cixel.c2 = printable[ncolors / MAXPRINTABLE];
	 *(Map + j*width + i) = cmap[ncolors].cixel ;  
	 if (ncolors < MaxCixel) ncolors++ ;
	 if (ncolors % MAXPRINTABLE) {        /*  => much memory for cmap */
	   cmap = (CmapEntry *) realloc((char*)cmap,
				(ncolors+MAXPRINTABLE)*sizeof(CmapEntry)); 
	   if (cmap == NULL) {
	     RETURN (PixmapNoMemory) ;
	   }
	 }
       }
     }
   }       
   if (ncolors == MaxCixel) ncolors++ ;
   cpp = (ncolors <= MAXPRINTABLE) ? 1 : 2;

   /* Write out standard header */
   fprintf(fstream, "#define %s_format %d\n",stripname,XPM_FORMAT);
   fprintf(fstream, "#define %s_width %d\n",stripname,width);
   fprintf(fstream, "#define %s_height %d\n",stripname,height);
   fprintf(fstream, "#define %s_ncolors %d\n",stripname,ncolors);
   fprintf(fstream, "#define %s_chars_per_pixel %d\n",stripname,cpp);

   fprintf(fstream, "static char * %s_colors[] = {\n",stripname);

   for (i=0; i<ncolors ; i++) {
     switch (cpp) {
     case 1:
       fprintf(fstream, "\"%c\" , \"#%04X%04X%04X\"",
	       cmap[i].cixel.c1,
	       cmap[i].red,cmap[i].green,cmap[i].blue);
       break;
     case 2:
       fprintf(fstream, "\"%c%c\" , \"#%04X%04X%04X\"",
	       cmap[i].cixel.c1,cmap[i].cixel.c2,
	       cmap[i].red,cmap[i].green,cmap[i].blue);
       break;
     }
     if (i != (ncolors - 1)) fprintf(fstream,",\n"); else
                             fprintf(fstream,"\n"); 
   }     
   fprintf(fstream, "} ;\n");
   fprintf(fstream, "static char * %s_pixels[] = {\n",stripname);

   for (j=0; j<height ; j++) {
     fprintf(fstream, "\"");
     for (i=0; i<width ; i++) 
       switch (cpp) {
       case 1:
         fprintf(fstream, "%c", (*(Map + j*width + i)).c1);
         break;
       case 2:
         fprintf(fstream, "%c%c", (*(Map + j*width + i)).c1,
	         (*(Map + j*width + i)).c2);
         break;
       }
     if (j != (height - 1)) fprintf(fstream,"\",\n"); else
                            fprintf(fstream,"\"\n"); 
   }

   if (fprintf(fstream, "} ;\n") == -1) {   /* enough disk space for it all? */
     RETURN (PixmapOpenFailed);
   }

   fclose(fstream);
   unlink(backupname);
   rename(filename, backupname);
   if (rename(tmpname, filename)) {
     RETURN (PixmapOpenFailed);
   }

   RETURN (PixmapSuccess) ;
}

/****[ UTILITIES ]************************************************************
 * following routines are used in XWritePixmapFile() function 
 *****************************************************************************/

static char *
StripName(name)
char *name;
{
	/* char	*begin = (char *) rindex(name, '/');*/
	char	*begin = (char *) strrchr(name, '/');
	char	*end, *result;
	int	length;

	begin = (begin ? begin+1 : name);
     /* change to rindex to allow longer names */	
	/*end = (char *) index(begin, '.');*/
	end = (char *) strchr(begin, '.');
	length = (end ? (end - begin) : strlen(begin));
	result = (char *) malloc(length + 1);
	if (result != NULL) {
		strncpy(result, begin, length);
		result[length] = '\0';
	}
	return (result);
}

static char *
BackupName(name)
char *name;
{
	int	length = name ? strlen(name) : 0;
	char	*result = (char *) malloc(length + 2);

	if (result != NULL) {
		strncpy(result, name, length);
		result[length] = '~';
		result[length+1] = '\0';
	}
	return (result);
}

static char *
TmpName(name)
char *name;
{
	/*char	*stop = (char *) rindex(name, '/');*/
	char	*stop = (char *) strrchr(name, '/');
	char	*tmp = "_tmp_pixmap_";
	int	pathname_length = (stop ? (stop - name + 1) : 0);
	int	tmp_length = strlen(tmp);
	int	result_length = pathname_length + tmp_length;
	char	*result = (char *) malloc(result_length + 1);

	if (result != NULL) {
		strncpy(result, name, pathname_length);
		strncpy(result+pathname_length, tmp, tmp_length);
		result[result_length] = '\0';
	}
	return (result);
}

/****<eof>********************************************************************/
